from epics import caget
import numpy as np


def readMatrix(basename, rows, cols, firstRow=1, firstCol=1, suffix=None):
    '''
    readMatrix(basename, rows, cols, firstRow=1, firstCol=1, suffix=None

    Reads epics channels from a EPICS channel matrix.

    Arguments
    ---------
    basename (str): Base name of EPICS channel matrix (See notes)
    rows (int): Number of rows to read
    cols (int): Number of columns to read
    firstRow (int): Index of the first row to read from the matrix [1]
    firstCol (int): Index of the first column to read from the matrix [1]
    suffix (str): Suffix after the matrix enumeration in channel names [None]

    Returns
    --------
    mat (numpy.ndarray): Numpy 2-D array containing the read matrix

    Notes
    --------
    The channel names in matrix would be constructed as:
        For i as row index, and j as column index:
            if suffix is None:
                basename_i_j
            else:
                basename_i_j_suffix
    '''
    mat = np.zeros((rows, cols))
    for ii in range(firstRow, firstRow + rows):
        for jj in range(firstCol, firstCol + cols):
            chName = basename + '_' + str(ii) + '_' + str(jj)
            if suffix is not None:
                chName = chName + '_' + suffix
            mat[ii - firstRow, jj-firstCol] = caget(chName)
    return mat



from epics import caput
import numpy as np
from .readMatrix import readMatrix
from time import sleep
from . import backUpVals, restoreMethods


def writeMatrix(basename, mat, firstRow=1, firstCol=1, suffix=None,
                tramp=None, bak=backUpVals):
    '''
    writeMatrix(basename, mat, firstRow=1, firstCol=1, suffix=None, tramp=None,
                bak=backUpVals)

    Write epics channels on a EPICS channel matrix.

    Arguments
    ---------
    basename (str): Base name of EPICS channel matrix (See notes)
    mat (numpy.ndarray): Numpy 2-D array matrix to write
    firstRow (int): Index of the first row to read from the matrix [1]
    firstCol (int): Index of the first column to read from the matrix [1]
    suffix (str): Suffix after the matrix enumeration in channel names [None]
    tramp (float): Ramp time for change in value [None]

    Returns
    --------
    retMat (numpy.ndarray): Numpy 2-D array containing output of each matrix
                            element write operation
                            1  on succesful completion
                            -1 or other negative value on failure of the
                            low-level CA put.
                            None  on a failure to connect to the PV.

    Notes
    --------
    The channel names in matrix would be constructed as:
        For i as row index, and j as column index:
            if suffix is None:
                basename_i_j
            else:
                basename_i_j_suffix
    '''
    curMat = readMatrix(basename=basename, rows=mat.shape[0],
                        cols=mat.shape[1], firstRow=firstRow,
                        firstCol=firstCol, suffix=suffix)
    if basename not in bak:
        bak[basename] = {'type': 'matrix', 'name': basename, 'suffix': suffix,
                         'value': curMat, 'firstRow': firstRow,
                         'firstCol': firstCol, 'tramp': tramp, 'sno': len(bak)}
    if tramp is not None:
        # Special case, changing filter gains, can use TRAMP
        rampYourself = True
        if suffix == 'GAIN':
            try:
                curTramp = readMatrix(basename=basename, rows=mat.shape[0],
                                      cols=mat.shape[1], firstRow=firstRow,
                                      firstCol=firstCol, suffix='TRAMP')
                for ii in range(firstRow, firstRow + np.shape(mat)[0]):
                    for jj in range(firstCol, firstCol + np.shape(mat)[1]):
                        chName = (basename + '_' + str(ii) + '_' + str(jj)
                                  + '_TRAMP')
                        caput(chName, tramp)
                rampYourself = False
            except BaseException:
                rampYourself = True
        if rampYourself:
            rampSteps = 100
            # Get current matrix values
            stepMat = (mat - curMat) / rampSteps
            for tstep in range(1, rampSteps):
                for ii in range(firstRow, firstRow + np.shape(mat)[0]):
                    for jj in range(firstCol, firstCol + np.shape(mat)[1]):
                        chName = basename + '_' + str(ii) + '_' + str(jj)
                        if suffix is not None:
                            chName = chName + '_' + suffix
                        matToWrite = (curMat[ii-firstRow, jj-firstCol]
                                      + tstep * stepMat[ii-firstRow,
                                                        jj-firstCol])
                        caput(chName, matToWrite)
                sleep(tramp/rampSteps)
    # Finally write the required matrix
    retMat = np.zeros_like(mat)
    for ii in range(firstRow, firstRow + np.shape(mat)[0]):
        for jj in range(firstCol, firstCol + np.shape(mat)[1]):
            chName = basename + '_' + str(ii) + '_' + str(jj)
            if suffix is not None:
                chName = chName + '_' + suffix
            matii = ii-firstRow
            matjj = jj-firstCol
            retMat[matii, matjj] = caput(chName, mat[matii, matjj])
    if tramp is not None:
        if not rampYourself:
            sleep(tramp + 0.5)    # Wait for ramping to end
            for ii in range(firstRow, firstRow + np.shape(mat)[0]):
                for jj in range(firstCol, firstCol + np.shape(mat)[1]):
                    chName = (basename + '_' + str(ii) + '_' + str(jj)
                              + '_TRAMP')
                    caput(chName, curTramp[ii-firstRow, jj-firstCol])
    return retMat


def restoreMatrix(bakVal):
    writeMatrix(basename=bakVal['name'], suffix=bakVal['suffix'],
                mat=bakVal['value'], firstCol=bakVal['firstCol'],
                firstRow=bakVal['firstRow'], tramp=bakVal['tramp'], bak={})


restoreMethods['matrix'] = restoreMatrix



import epics
from . import backUpVals, restoreMethods
from time import sleep


def caput(pvname, value, wait=False, timeout=60.0, tramp=None, bak=backUpVals):
    """
    caput(pvname, value, wait=False, timeout=60.0, bak=backUpVals, tramp=None)

    Put a value to an epics Process Variable (PV) with added option of
    ramping the value lineary. This function stores a backup of the channel
    before changing it that will be reverted by restoreEpics() call.

    Arguments
    ---------
    pvname (str): name of PV
    value (any): value to put to PV
    wait (bool): whether to wait for processing to complete [False]
    timeout (float): maximum time to wait for the processing to complete [60]
    tramp (float): Ramp time for change in value [None]
    bak (dict): Dictionary of backup values during a particular instance

    Returns
    --------
    1  on succesful completion
    -1 or other negative value on failure of the low-level CA put.
    None  on a failure to connect to the PV.

    Examples
    ---------
    To put a value to a PV and return as soon as possible:
    >>> caput('xx', 3.0)
    To wait for processing to finish, use 'wait=True':
    >>> caput('xx', 3.0, wait=True)
    To ramp the change in value linearily in time and wait for it to happen:
    >>> caput('xx', 3.0, wait=True, timeout=5)
    """
    val = epics.caget(pvname, timeout=timeout)
    if pvname not in bak:
        bak[pvname] = {'type': 'channel', 'value': val, 'tramp': tramp,
                       'sno': len(bak), 'name': pvname}
    if tramp is None:
        retVal = epics.caput(pvname=pvname, value=value, wait=wait,
                             timeout=timeout)
    else:
        rampYourself = True
        fus = pvname.rfind('_') + 1
        # If this looks like a GAIN channel, look for TRAMP channel
        if pvname[fus:] in ['GAIN', 'OFFSET']:
            trch = pvname[:-4] + 'TRAMP'
            tbak = epics.caget(pvname=trch, timeout=timeout)
            if tbak is not None:
                rampYourself = False
            else:
                rampYourself = True
        if not rampYourself:
            epics.caput(pvname=trch, value=tramp, wait=wait, timeout=timeout)
            retVal = epics.caput(pvname=pvname, value=value, wait=wait,
                                 timeout=timeout)
            if wait:
                sleep(tramp)
            epics.caput(pvname=trch, value=tbak, wait=wait, timeout=timeout)
        else:
            rampSteps = 100
            valStep = (value - val) / rampSteps
            for tstep in range(1, rampSteps):
                epics.caput(pvname=pvname, value=val + valStep * tstep,
                            wait=wait, timeout=timeout)
                sleep(tramp / rampSteps)
            retVal = epics.caput(pvname=pvname, value=value, wait=wait,
                                 timeout=timeout)
    return retVal


def restoreChannel(bakVal):
    epics.caput(bakVal['name'], bakVal['value'])


restoreMethods['channel'] = restoreChannel



from . import backUpVals, restoreMethods


def restoreEpics(bak=backUpVals, restoreMethods=restoreMethods):
    '''
    restoreEpics(bak=backUpVals)

    Restores EPICS channels to their backed up values using predefind
    restoreMethods.

    Arguments
    ---------
    bak (dict): Dictionary of backup values during a particular instance
                [restoreEpics.backUpVals]
    restoreMethods (dict): Dictionary containing restortion methods for
                           different kinds of backed up values
                           [restoreEpics.restoreMethods]

    Notes
    --------
    bak dictionary minimum required format:
      * The key value of the dictionary does not matter.
      * Each element must be a dictionary containing atleast:
        * 'sno' key which is uniquely defined to store chronology of changes.
        * 'type' key matching a restoreMethods type to select the appropriate
          restoration method.
    restoreMethods dictionary minimum required format:
      * The keys of this dictionary should correspond to the type of object
        they would restore eg channel, matrix etc.
      * The value of the dictionary should be a method that takes in entire
        backed up element dictionary as argument and restores that element.

    Examples
    --------
    Use without arguments to restore everything that was changed using methods
    of this package:
    >> restoreEpics()
    If you want to restore a particular set of backed up values only, provide
    them as a dictionary:
    >> restoreEpics(bak=bak)
    If you want to use new custom restoration methods, define them in a
    dictionary and pass them along
    >> restoreEpics(bak=bak, restoreMethods=customMethods)
    '''
    print('Restoring channel values...')
    for ii in range(len(bak)-1, -1, -1):
        for ele in bak.values():
            if ii == ele['sno']:
                restoreMethods[ele['type']](ele)
